"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const precise_proofs_js_1 = require("precise-proofs-js");
const utils_general_1 = require("@energyweb/utils-general");
const PreciseProofEntity_1 = require("./PreciseProofEntity");
const origin_backend_core_1 = require("@energyweb/origin-backend-core");
const NULL_ADDRESS = '0x0000000000000000000000000000000000000000';
const NULL_HASH = '0x0000000000000000000000000000000000000000000000000000000000000000';
class Entity extends PreciseProofEntity_1.PreciseProofEntity {
    constructor(id, configuration) {
        super(id, configuration);
        this.initialized = false;
        this.claimedShares = {};
        this.ownedShares = {};
        this.ownedSharesPrivate = {};
    }
    async sync() {
        if (this.id === null) {
            return this;
        }
        const registry = this.configuration.blockchainProperties.registry;
        const certOnChain = await registry.getCertificate(this.id);
        this.data = certOnChain.data;
        const issuer = this.configuration.blockchainProperties.issuer;
        const decodedData = await issuer.decodeData(this.data);
        const allIssuanceEvents = await registry.getAllIssuanceSingleEvents({
            filter: { _id: this.id }
        });
        const creationBlock = await this.configuration.blockchainProperties.web3.eth.getBlock(allIssuanceEvents[0].blockNumber);
        this.generationStartTime = Number(decodedData['0']);
        this.generationEndTime = Number(decodedData['1']);
        this.deviceId = decodedData['2'];
        this.issuer = certOnChain.issuer;
        this.creationTime = Number(creationBlock.timestamp);
        this.creationBlockHash = creationBlock.hash;
        this.certificationRequestId = Number(await issuer.getCertificationRequestIdForCertificate(this.id));
        this.claimedShares = await this.calculateClaims();
        this.ownedShares = await this.calculateOwnership();
        this.propertiesDocumentHash = await issuer.getCertificateCommitment(this.id);
        if (this.propertiesDocumentHash !== NULL_HASH) {
            const { commitment } = await this.getCommitment();
            this.ownedSharesPrivate = commitment !== null && commitment !== void 0 ? commitment : {};
        }
        const getShareSum = (shares) => Object.keys(shares).map(owner => shares[owner]).reduce((a, b) => a + b, 0);
        this.energy = getShareSum(this.ownedShares) + getShareSum(this.claimedShares);
        this.initialized = true;
        if (this.configuration.logger) {
            this.configuration.logger.verbose(`Certificate ${this.id} synced`);
        }
        return this;
    }
    get owners() {
        var _a, _b, _c;
        const owners = {};
        const allOwnerAddresses = [...new Set([
                ...Object.keys(this.ownedShares),
                ...Object.keys(this.ownedSharesPrivate),
                ...Object.keys(this.claimedShares)
            ])];
        for (const ownerAddress of allOwnerAddresses) {
            if (!owners[ownerAddress]) {
                owners[ownerAddress] = {
                    owned: 0,
                    ownedPrivate: 0,
                    claimed: 0
                };
            }
            owners[ownerAddress].owned += (_a = this.ownedShares[ownerAddress]) !== null && _a !== void 0 ? _a : 0;
            owners[ownerAddress].ownedPrivate += (_b = this.ownedSharesPrivate[ownerAddress]) !== null && _b !== void 0 ? _b : 0;
            owners[ownerAddress].claimed += (_c = this.claimedShares[ownerAddress]) !== null && _c !== void 0 ? _c : 0;
        }
        return owners;
    }
    async claim(amount) {
        const { publicVolume, privateVolume } = this.ownedVolume();
        if (publicVolume === 0) {
            throw new Error(privateVolume === 0
                ? `claim(): Unable to claim certificate. You do not own a share in the certificate.`
                : `claim(): Can't claim private volumes. Please migrate some volume to public first.`);
        }
        if (amount && amount > publicVolume) {
            const totalOwned = publicVolume + privateVolume;
            throw new Error(`claim(): Can't claim ${amount} Wh. ${totalOwned < amount
                ? `You only own ${publicVolume} Wh.`
                : `Please migrate ${amount - publicVolume} Wh from private to public.`}`);
        }
        const { randomHex, hexToBytes } = this.configuration.blockchainProperties.web3.utils;
        // TO-DO: replace with proper claim data
        const claimData = hexToBytes(randomHex(32));
        const registry = this.configuration.blockchainProperties.registry;
        const owner = this.configuration.blockchainProperties.activeUser.address;
        return registry.safeTransferAndClaimFrom(owner, owner, this.id, amount || publicVolume, this.data, claimData, utils_general_1.Configuration.getAccount(this.configuration));
    }
    async requestMigrateToPublic() {
        const { privateVolume } = this.ownedVolume();
        if (privateVolume === 0) {
            throw new Error('migrateToPublic(): No private volume owned.');
        }
        const owner = this.configuration.blockchainProperties.activeUser.address.toLowerCase();
        const issuer = this.configuration.blockchainProperties.issuer;
        const { salts } = await this.getCommitment();
        const calculatedOffChainStorageProperties = this.generateAndAddProofs(this.ownedSharesPrivate, salts);
        const ownerAddressLeafHash = calculatedOffChainStorageProperties.leafs.find(leaf => leaf.key === owner).hash;
        return issuer.requestMigrateToPublic(this.id, ownerAddressLeafHash, utils_general_1.Configuration.getAccount(this.configuration));
    }
    async migrateToPublic() {
        const issuer = this.configuration.blockchainProperties.issuer;
        const migrationRequestId = await issuer.getMigrationRequestId(this.id, utils_general_1.Configuration.getAccount(this.configuration));
        const migrationRequest = await issuer.getMigrationRequest(migrationRequestId, utils_general_1.Configuration.getAccount(this.configuration));
        const requestor = migrationRequest.owner.toLowerCase();
        const { privateVolume } = this.ownedVolume(requestor);
        if (privateVolume === 0) {
            throw new Error(`migrateToPublic(): Requestor doesn't own any private volume in certificate #${this.id}.`);
        }
        const { salts } = await this.getCommitment();
        const calculatedOffChainStorageProperties = this.generateAndAddProofs(this.ownedSharesPrivate, salts);
        const theProof = precise_proofs_js_1.PreciseProofs.createProof(requestor, calculatedOffChainStorageProperties.leafs, false);
        const onChainProof = theProof.proofPath.map(p => ({
            left: !!p.left,
            hash: p.left || p.right
        }));
        const { salt } = theProof;
        await issuer.migrateToPublic(migrationRequestId, privateVolume, salt, onChainProof, utils_general_1.Configuration.getAccount(this.configuration));
    }
    async transfer(to, amount, privately = false) {
        var _a;
        const fromAddress = this.configuration.blockchainProperties.activeUser.address.toLowerCase();
        const toAddress = to.toLowerCase();
        const issuer = this.configuration.blockchainProperties.issuer;
        const { publicVolume, privateVolume } = this.ownedVolume();
        const availableAmount = (privately ? privateVolume : publicVolume);
        amount = amount !== null && amount !== void 0 ? amount : availableAmount;
        if (amount === 0 || amount > availableAmount) {
            throw new Error(`transfer(): unable to send amount ${amount} Wh. Sender ${fromAddress} has a balance of ${publicVolume + privateVolume} Wh (public: ${publicVolume}, private: ${privateVolume})`);
        }
        if (privately) {
            const proposedOwnerShares = { ...this.ownedSharesPrivate };
            proposedOwnerShares[fromAddress] -= amount;
            proposedOwnerShares[toAddress] = ((_a = proposedOwnerShares[toAddress]) !== null && _a !== void 0 ? _a : 0) + amount;
            const commitmentProof = this.generateAndAddProofs(proposedOwnerShares);
            const ownerAddressLeafHash = commitmentProof.leafs.find(leaf => leaf.key === fromAddress).hash;
            const tx = await issuer.requestPrivateTransfer(this.id, ownerAddressLeafHash, utils_general_1.Configuration.getAccount(this.configuration));
            return this.saveCommitment({
                ...commitmentProof,
                txHash: tx.transactionHash
            });
        }
        const registry = this.configuration.blockchainProperties.registry;
        return registry.safeTransferFrom(fromAddress, toAddress, this.id, amount, this.data, utils_general_1.Configuration.getAccount(this.configuration));
    }
    async approvePrivateTransfer() {
        const issuer = this.configuration.blockchainProperties.issuer;
        const previousCommitment = this.propertiesDocumentHash;
        const newCommitmentProof = await this.getPendingTransferCommitment();
        const request = await issuer.getPrivateTransferRequest(this.id, utils_general_1.Configuration.getAccount(this.configuration));
        if (!request) {
            throw new Error(`approvePrivateTransfer(): no pending requests to approve.`);
        }
        const theProof = precise_proofs_js_1.PreciseProofs.createProof(request.owner.toLowerCase(), newCommitmentProof.leafs, false);
        const onChainProof = theProof.proofPath.map(p => ({
            left: !!p.left,
            hash: p.left || p.right
        }));
        this.propertiesDocumentHash = newCommitmentProof.rootHash;
        const tx = issuer.approvePrivateTransfer(this.id, onChainProof, previousCommitment, newCommitmentProof.rootHash, utils_general_1.Configuration.getAccount(this.configuration));
        await this.certificateClient.approvePendingOwnershipCommitment(this.id);
        return tx;
    }
    async revoke() {
        const issuer = this.configuration.blockchainProperties.issuer;
        return issuer.revokeCertificate(this.id, utils_general_1.Configuration.getAccount(this.configuration));
    }
    async getAllCertificateEvents() {
        return exports.getAllCertificateEvents(this.id, this.configuration);
    }
    isOwned(byAddress) {
        const { publicVolume, privateVolume } = this.ownedVolume(byAddress);
        return (publicVolume + privateVolume) > 0;
    }
    ownedVolume(byAddress) {
        var _a, _b;
        const owner = byAddress !== null && byAddress !== void 0 ? byAddress : this.configuration.blockchainProperties.activeUser.address;
        const ownerBalances = this.owners[owner.toLowerCase()];
        return {
            publicVolume: ((_a = ownerBalances === null || ownerBalances === void 0 ? void 0 : ownerBalances.owned) !== null && _a !== void 0 ? _a : 0) - this.claimedVolume(owner),
            privateVolume: (_b = ownerBalances === null || ownerBalances === void 0 ? void 0 : ownerBalances.ownedPrivate) !== null && _b !== void 0 ? _b : 0
        };
    }
    isClaimed(byAddress) {
        const claimedVolume = this.claimedVolume(byAddress);
        return claimedVolume > 0;
    }
    claimedVolume(byAddress) {
        var _a;
        const owner = byAddress !== null && byAddress !== void 0 ? byAddress : this.configuration.blockchainProperties.activeUser.address;
        return (_a = this.claimedShares[owner.toLowerCase()]) !== null && _a !== void 0 ? _a : 0;
    }
    async calculateOwnership() {
        var _a;
        const ownedShares = {};
        const registry = this.configuration.blockchainProperties.registry;
        let transferSingleEvents = (await registry.getAllTransferSingleEvents({
            filter: { _id: this.id.toString() }
        }))
            .filter(e => e.returnValues._id === this.id.toString())
            .map(e => e.returnValues);
        const transferBatchEvents = (await registry.getAllTransferBatchEvents()).map(e => e.returnValues);
        // Convert TransferBatch to TransferSingle event
        for (const event of transferBatchEvents.filter(e => e._ids.includes(this.id.toString()))) {
            for (let i = 0; i < event._ids.length; i++) {
                if (event._ids[i] === this.id.toString()) {
                    transferSingleEvents.push({
                        _id: event._ids[i],
                        _operator: event._operator,
                        _to: event._to,
                        _from: event._from,
                        _value: event._values[i]
                    });
                }
            }
        }
        for (const event of transferSingleEvents) {
            const { _from, _to, _value } = event;
            const fromAddress = _from.toLowerCase();
            const toAddress = _to.toLowerCase();
            const valueTransferred = Number(_value);
            if (_from !== NULL_ADDRESS) {
                ownedShares[fromAddress] = ownedShares[fromAddress] - valueTransferred;
            }
            if (ownedShares[toAddress] === null || ownedShares[toAddress] === undefined) {
                ownedShares[toAddress] = 0;
            }
            ownedShares[toAddress] += valueTransferred;
        }
        // Validate if all balances have been correctly calculated
        for (const ownerAddress of Object.keys(ownedShares)) {
            const ownedBalance = Number(await registry.balanceOf(ownerAddress, this.id));
            const calculatedBalance = ownedShares[ownerAddress] - ((_a = this.claimedShares[ownerAddress]) !== null && _a !== void 0 ? _a : 0);
            if (ownedBalance != calculatedBalance) {
                throw new Error(`Non-matching owned balances. Please re-sync the certificate data.\nCalculated from events: ${calculatedBalance}\nRegistry.balanceOf(): ${ownedBalance}`);
            }
        }
        return ownedShares;
    }
    async calculateClaims() {
        const claimedShares = {};
        const registry = this.configuration.blockchainProperties.registry;
        let claimSingleEvents = (await registry.getAllClaimSingleEvents({
            filter: { _id: this.id.toString() }
        }))
            .filter(e => e.returnValues._id === this.id.toString())
            .map(e => e.returnValues);
        const claimBatchEvents = (await registry.getAllClaimBatchEvents()).map(e => e.returnValues);
        // Convert ClaimBatch to ClaimSingle event
        for (const event of claimBatchEvents.filter(e => e._ids.includes(this.id.toString()))) {
            for (let i = 0; i < event._ids.length; i++) {
                if (event._ids[i] === this.id.toString()) {
                    claimSingleEvents.push({
                        _claimIssuer: event._claimIssuer,
                        _claimSubject: event._claimSubject,
                        _topic: event._topics[i],
                        _id: event._ids[i],
                        _value: event._values[i],
                        _claimData: event._claimData[i]
                    });
                }
            }
        }
        for (const event of claimSingleEvents) {
            const { _claimSubject, _value } = event;
            const claimSubject = _claimSubject.toLowerCase();
            const valueClaimed = Number(_value);
            if (claimedShares[claimSubject] === null || claimedShares[claimSubject] === undefined) {
                claimedShares[claimSubject] = 0;
            }
            claimedShares[claimSubject] += valueClaimed;
        }
        // Validate if all balances have been correctly calculated
        for (const ownerAddress of Object.keys(claimedShares)) {
            const claimedBalance = await registry.claimedBalanceOf(ownerAddress, this.id);
            if (claimedBalance != claimedShares[ownerAddress]) {
                throw new Error(`Non-matching claimed balances. Please re-sync the certificate data.\nCalculated from events: ${claimedShares[ownerAddress]}\nRegistry.balanceOf(): ${claimedBalance}`);
            }
        }
        return claimedShares;
    }
}
exports.Entity = Entity;
exports.createCertificate = async (to, value, fromTime, toTime, deviceId, configuration, isVolumePrivate = false) => {
    if (value > origin_backend_core_1.MAX_ENERGY_PER_CERTIFICATE) {
        throw new Error(`Too much energy requested. Requested: ${value}, Max: ${origin_backend_core_1.MAX_ENERGY_PER_CERTIFICATE}`);
    }
    const newEntity = new Entity(null, configuration);
    const getIdFromLogs = (logs) => configuration.blockchainProperties.web3.utils.hexToNumber(logs[0].topics[2]);
    const issuer = configuration.blockchainProperties.issuer;
    const data = await issuer.encodeData(fromTime, toTime, deviceId);
    let tx;
    if (isVolumePrivate) {
        const ownershipCommitment = {
            [to.toLowerCase()]: value
        };
        const commitmentProof = newEntity.generateAndAddProofs(ownershipCommitment);
        tx = await issuer.issuePrivate(to, commitmentProof.rootHash, data, utils_general_1.Configuration.getAccount(configuration));
        newEntity.id = getIdFromLogs(tx.logs);
        newEntity.propertiesDocumentHash = commitmentProof.rootHash;
        await newEntity.saveCommitment({
            ...commitmentProof,
            txHash: tx.transactionHash
        });
    }
    else {
        tx = await issuer.issue(to, value, data, utils_general_1.Configuration.getAccount(configuration));
        newEntity.id = getIdFromLogs(tx.logs);
    }
    if (configuration.logger) {
        configuration.logger.info(`Certificate ${newEntity.id} created`);
    }
    return newEntity.sync();
};
async function claimCertificates(certificateIds, configuration) {
    const certificatesPromises = certificateIds.map(certId => new Entity(certId, configuration).sync());
    const certificates = await Promise.all(certificatesPromises);
    const owned = certificates.map(cert => cert.isOwned());
    const ownsAllCertificates = owned.every(isOwned => isOwned === true);
    if (!ownsAllCertificates) {
        throw new Error(`You can only claim your own certificates`);
    }
    const values = certificates.map(cert => cert.ownedVolume().publicVolume);
    const { randomHex, hexToBytes } = configuration.blockchainProperties.web3.utils;
    // TO-DO: replace with proper claim data
    const claimData = certificates.map(cert => hexToBytes(randomHex(32)));
    const data = hexToBytes(randomHex(32));
    const { from } = utils_general_1.Configuration.getAccount(configuration);
    return configuration.blockchainProperties.registry.safeBatchTransferAndClaimFrom(from, from, certificateIds, values, data, claimData, utils_general_1.Configuration.getAccount(configuration));
}
exports.claimCertificates = claimCertificates;
async function transferCertificates(certificateIds, to, configuration) {
    const certificatesPromises = certificateIds.map(certId => new Entity(certId, configuration).sync());
    const certificates = await Promise.all(certificatesPromises);
    const values = certificates.map(cert => cert.ownedVolume().publicVolume);
    const { randomHex, hexToBytes } = configuration.blockchainProperties.web3.utils;
    // TO-DO: replace with proper data
    const data = hexToBytes(randomHex(32));
    const { from } = utils_general_1.Configuration.getAccount(configuration);
    return configuration.blockchainProperties.registry.safeBatchTransferFrom(from, to, certificateIds, values, data, utils_general_1.Configuration.getAccount(configuration));
}
exports.transferCertificates = transferCertificates;
async function getAllCertificates(configuration) {
    const issuer = configuration.blockchainProperties.issuer;
    const totalRequests = await issuer.totalRequests();
    const certificatePromises = Array(Number(totalRequests))
        .fill(null)
        .map(async (item, index) => {
        const certId = await issuer.getCertificateIdForCertificationRequest(index + 1);
        return certId > 0 ? new Entity(certId, configuration).sync() : null;
    });
    const certificates = await Promise.all(certificatePromises);
    return certificates.filter(cert => cert !== null);
}
exports.getAllCertificates = getAllCertificates;
exports.getAllCertificateEvents = async (certId, configuration) => {
    const registry = configuration.blockchainProperties.registry;
    const allEvents = await registry.getAllEvents({
        topics: [
            null,
            configuration.blockchainProperties.web3.utils.padLeft(configuration.blockchainProperties.web3.utils.fromDecimal(certId), 64, '0')
        ],
        fromBlock: 0,
        toBlock: 'latest'
    });
    const returnEvents = [];
    for (const fullEvent of allEvents) {
        // we have to remove some false positives due to ERC721 interface
        if (fullEvent.event === 'Transfer') {
            if (fullEvent.returnValues.tokenId === `${certId}`) {
                returnEvents.push(fullEvent);
            }
        }
        else {
            returnEvents.push(fullEvent);
        }
    }
    // we also have to search
    if (certId !== 0) {
        const transferEvents = await registry.getAllTransferSingleEvents({
            topics: [
                '0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef',
                null,
                null,
                configuration.blockchainProperties.web3.utils.padLeft(configuration.blockchainProperties.web3.utils.fromDecimal(certId), 64, '0')
            ],
            fromBlock: 0,
            toBlock: 'latest'
        });
        for (const transferEvent of transferEvents) {
            returnEvents.push(transferEvent);
        }
    }
    return returnEvents;
};
//# sourceMappingURL=Certificate.js.map